/*
    PPNavigatorPopupPanelController.m

    Copyright 2013-2016 Josh Freeman
    http://www.twilightedge.com

    This file is part of PikoPixel for Mac OS X and GNUstep.
    PikoPixel is a graphical application for drawing & editing pixel-art images.

    PikoPixel is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3,
    as published by the Free Software Foundation.

    PikoPixel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#import "PPNavigatorPopupPanelController.h"

#import "PPDocument.h"
#import "PPNavigatorPopupView.h"
#import "PPDocumentWindowController.h"
#import "PPCanvasView.h"
#import "PPUIColors_Panels.h"
#import "PPCursorManager.h"
#import "NSObject_PPUtilities.h"


#define kNavigatorPopupPanelNibName  @"NavigatorPopupPanel"


@interface PPNavigatorPopupPanelController (PrivateMethods)

- (void) addAsObserverForPPCanvasViewNotifications;
- (void) removeAsObserverForPPCanvasViewNotifications;
- (void) handlePPCanvasViewNotification_ChangedZoomFactor: (NSNotification *) notification;

- (void) handlePPDocumentNotification_UpdatedMergedVisibleThumbnailImage:
                                                            (NSNotification *) notification;
- (void) handlePPDocumentNotification_UpdatedDrawingLayerThumbnailImage:
                                                            (NSNotification *) notification;
- (void) handlePPDocumentNotification_UpdatedBackgroundSettings:
                                                            (NSNotification *) notification;

- (void) addAsObserverForPPDocumentWindowControllerNotifications;
- (void) removeAsObserverForPPDocumentWindowControllerNotifications;
- (void) handlePPDocumentWindowControllerNotification_ChangedCanvasDisplayMode:
                                                                (NSNotification *) notification;

- (void) setupWindowControllerForCurrentPPDocument;
- (void) setupWithCanvasView: (PPCanvasView *) canvasView;
- (void) updateZoomSliderPosition;
- (void) updateNavigatorImage;
- (void) updateNavigatorBackground;

- (void) disableNavigatorViewMouseTracking;
- (void) enableNavigatorViewMouseTracking;

@end

@implementation PPNavigatorPopupPanelController

#pragma mark Actions

- (IBAction) zoomButtonPressed: (id) sender
{
    if (sender == _decreaseZoomButton)
    {
        [_canvasView decreaseZoomFactor];
    }
    else if (sender == _increaseZoomButton)
    {
        [_canvasView increaseZoomFactor];
    }
}

- (IBAction) zoomSliderMoved: (id) sender
{
    NSAutoreleasePool *autoreleasePool;

    [self disableNavigatorViewMouseTracking];
    [[PPCursorManager sharedManager] setCursor: [NSCursor arrowCursor]
                                        atLevel: kPPCursorLevel_PopupPanel
                                        isDraggingMouse: YES];

    autoreleasePool = [[NSAutoreleasePool alloc] init];

    _disallowUpdatesToZoomSliderPosition = YES;

    [_canvasView setZoomFactor: [_zoomSlider floatValue]];

    _disallowUpdatesToZoomSliderPosition = NO;

    [autoreleasePool release];

    [self ppPerformSelectorAtomicallyFromNewStackFrame:
                                                @selector(enableNavigatorViewMouseTracking)];
}

#pragma mark NSWindowController overrides

- (void) windowDidLoad
{
    _navigatorDisplayMode = kPPLayerDisplayMode_VisibleLayers;

    [_zoomSlider setMinValue: kMinCanvasZoomFactor];
    [_zoomSlider setMaxValue: kMaxCanvasZoomFactor];

    [super windowDidLoad];
}

#pragma mark PPPopupPanelController overrides

+ (NSString *) panelNibName
{
    return kNavigatorPopupPanelNibName;
}

- (NSColor *) backgroundColorForPopupPanel
{
    return kUIColor_NavigatorPopupPanel_Background;
}

- (void) setupPanelForCurrentPPDocument
{
    [super setupPanelForCurrentPPDocument];

    [self setupWindowControllerForCurrentPPDocument];

    _navigatorDisplayMode = [_ppDocumentWindowController canvasDisplayMode];

    [self setupWithCanvasView: [_ppDocumentWindowController canvasView]];
    [self updateNavigatorImage];
    [self updateNavigatorBackground];
}

- (void) setupPanelAfterVisibilityChange
{
    [_navigatorView setupMouseTracking];

    if (![self panelIsVisible])
    {
        if ([_navigatorView mouseIsDraggingVisibleBounds])
        {
            [_navigatorView forceStopDraggingVisibleBounds];
        }

        // if the navigator popup's controls were clicked, sometimes the canvas view won't
        // receive a mouseEntered event after the popup hides, leaving the wrong cursor, so
        // force the canvas view to update its cursor:
        [_canvasView updateCursorForCurrentMouseLocation];
    }

    [super setupPanelAfterVisibilityChange];
}

- (void) handleDirectionCommand: (PPDirectionType) directionType
{
    switch (directionType)
    {
        case kPPDirectionType_Right:
        case kPPDirectionType_Up:
        {
            [_canvasView increaseZoomFactor];
        }
        break;

        case kPPDirectionType_Left:
        case kPPDirectionType_Down:
        {
            [_canvasView decreaseZoomFactor];
        }
        break;

        default:
        break;
    }
}

#pragma mark PPCanvasView notifications

- (void) addAsObserverForPPCanvasViewNotifications
{
    if (!_canvasView)
        return;

    [[NSNotificationCenter defaultCenter]
                            addObserver: self
                            selector:
                                @selector(handlePPCanvasViewNotification_ChangedZoomFactor:)
                            name: PPCanvasViewNotification_ChangedZoomFactor
                            object: _canvasView];
}

- (void) removeAsObserverForPPCanvasViewNotifications
{
    [[NSNotificationCenter defaultCenter] removeObserver: self
                                            name: PPCanvasViewNotification_ChangedZoomFactor
                                            object: _canvasView];
}

- (void) handlePPCanvasViewNotification_ChangedZoomFactor: (NSNotification *) notification
{
    [self updateZoomSliderPosition];
}

#pragma mark PPDocument notifications

- (void) addAsObserverForPPDocumentNotifications
{
    NSNotificationCenter *notificationCenter = [NSNotificationCenter defaultCenter];

    if (!_ppDocument)
        return;

    [notificationCenter addObserver: self
                        selector:
                            @selector(
                            handlePPDocumentNotification_UpdatedMergedVisibleThumbnailImage:)
                        name: PPDocumentNotification_UpdatedMergedVisibleThumbnailImage
                        object: _ppDocument];

    [notificationCenter addObserver: self
                        selector:
                            @selector(
                            handlePPDocumentNotification_UpdatedDrawingLayerThumbnailImage:)
                        name: PPDocumentNotification_UpdatedDrawingLayerThumbnailImage
                        object: _ppDocument];

    [notificationCenter addObserver: self
                        selector:
                            @selector(handlePPDocumentNotification_UpdatedBackgroundSettings:)
                        name: PPDocumentNotification_UpdatedBackgroundSettings
                        object: _ppDocument];
}

- (void) removeAsObserverForPPDocumentNotifications
{
    NSNotificationCenter *notificationCenter = [NSNotificationCenter defaultCenter];

    [notificationCenter removeObserver: self
                        name: PPDocumentNotification_UpdatedMergedVisibleThumbnailImage
                        object: _ppDocument];

    [notificationCenter removeObserver: self
                        name: PPDocumentNotification_UpdatedDrawingLayerThumbnailImage
                        object: _ppDocument];

    [notificationCenter removeObserver: self
                        name: PPDocumentNotification_UpdatedBackgroundSettings
                        object: _ppDocument];
}

- (void) handlePPDocumentNotification_UpdatedMergedVisibleThumbnailImage:
                                                            (NSNotification *) notification
{
    if (_navigatorDisplayMode == kPPLayerDisplayMode_VisibleLayers)
    {
        [self updateNavigatorImage];
    }
}

- (void) handlePPDocumentNotification_UpdatedDrawingLayerThumbnailImage:
                                                                (NSNotification *) notification
{
    if (_navigatorDisplayMode == kPPLayerDisplayMode_DrawingLayerOnly)
    {
        [self updateNavigatorImage];
    }
}

- (void) handlePPDocumentNotification_UpdatedBackgroundSettings:
                                                            (NSNotification *) notification
{
    [self updateNavigatorBackground];
}

#pragma mark PPDocumentWindowController notifications

- (void) addAsObserverForPPDocumentWindowControllerNotifications
{
    if (!_ppDocumentWindowController)
        return;

    [[NSNotificationCenter defaultCenter]
        addObserver: self
        selector:
            @selector(handlePPDocumentWindowControllerNotification_ChangedCanvasDisplayMode:)
        name: PPDocumentWindowControllerNotification_ChangedCanvasDisplayMode
        object: _ppDocumentWindowController];
}

- (void) removeAsObserverForPPDocumentWindowControllerNotifications
{
    [[NSNotificationCenter defaultCenter]
                        removeObserver: self
                        name: PPDocumentWindowControllerNotification_ChangedCanvasDisplayMode
                        object: _ppDocumentWindowController];
}

- (void) handlePPDocumentWindowControllerNotification_ChangedCanvasDisplayMode:
                                                                (NSNotification *) notification
{
    _navigatorDisplayMode = [_ppDocumentWindowController canvasDisplayMode];

    [self updateNavigatorImage];
}

#pragma mark Private methods

- (void) setupWindowControllerForCurrentPPDocument
{
    if (_ppDocumentWindowController)
    {
        [self removeAsObserverForPPDocumentWindowControllerNotifications];
    }

    _ppDocumentWindowController = [_ppDocument ppDocumentWindowController];

    if (_ppDocumentWindowController)
    {
        [self addAsObserverForPPDocumentWindowControllerNotifications];
    }
}

- (void) setupWithCanvasView: (PPCanvasView *) canvasView
{
    if (_canvasView)
    {
        [self removeAsObserverForPPCanvasViewNotifications];
        [_canvasView release];
    }

    _canvasView = [canvasView retain];

    if (_canvasView)
    {
        [self addAsObserverForPPCanvasViewNotifications];
    }

    [_navigatorView setCanvasView: canvasView];

    [self updateZoomSliderPosition];
}

- (void) updateZoomSliderPosition
{
    if (_disallowUpdatesToZoomSliderPosition)
        return;

    [_zoomSlider setFloatValue: [_canvasView zoomFactor]];
}

- (void) updateNavigatorImage
{
    NSImage *navigatorImage;

    if (_navigatorDisplayMode == kPPLayerDisplayMode_DrawingLayerOnly)
    {
        navigatorImage = [_ppDocument dissolvedDrawingLayerThumbnailImage];
    }
    else
    {
        navigatorImage = [_ppDocument mergedVisibleLayersThumbnailImage];
    }

    [_navigatorView setImage: navigatorImage];
}

- (void) updateNavigatorBackground
{
    NSImage *backgroundImage;

    backgroundImage = ([_ppDocument shouldDisplayBackgroundImage]) ?
                            [_ppDocument backgroundImage] : nil;

    [_navigatorView setBackgroundPattern: [_ppDocument backgroundPattern]
                    andBackgroundImage: backgroundImage];
}

- (void) disableNavigatorViewMouseTracking
{
    [_navigatorView disableMouseTracking: YES];
}

- (void) enableNavigatorViewMouseTracking
{
    [_navigatorView disableMouseTracking: NO];
}

@end
